
#include "gfx.h"

#define __PROG_TYPES_COMPAT__
#include <avr/pgmspace.h>
#include <avr/io.h>
#include <stdlib.h>
#include <util/delay.h>

#define HEIGHT	131
#define WIDTH	130
#define INVADER_SIZE 11
#define HEIGHT_OFF 3
#define WIDTH_OFF 2
int8_t INVADER_LN = 3;
int8_t INVADER_COL  =10;
#define LINE 3
#define COL 10
#define START_INV_X 20
#define START_INV_Y (WIDTH_OFF + INVADER_SIZE/2)
#define WIDTH_LIMIT (WIDTH - 1)
#define SHIP_HEIGHT 17
#define SHIP_WIDTH 17
#define SHIP_START_X 120
#define SHIP_START_Y WIDTH/2
#define START_HEALTH 10 
#define SHIP_HEIGHT 17
#define SHIP_WIDTH 17
#define BOSS_SPEED 2

static int8_t invader_shape[100] = {	0,1,0,0,0,0,0,0,0,1,0,
										0,1,1,1,0,0,0,1,1,1,0,
										0,0,0,1,1,1,1,1,0,0,0,
										0,0,1,1,0,1,0,1,1,0,0,
										0,1,1,1,1,1,1,1,1,1,0,
										0,0,1,1,1,1,1,1,1,0,0,
										0,0,0,1,1,1,1,1,0,0,0,
										0,0,0,1,0,1,0,1,0,0,0,
										0,0,1,0,0,1,0,0,1,0,0};
										
static int8_t ship_shape[SHIP_HEIGHT][SHIP_WIDTH] = {	{0,0,0,0,0,3,0,0,0,0,0,3,0,0,0,0,0},
														{0,0,0,0,0,3,0,0,3,0,0,3,0,0,0,0,0},
														{0,0,0,0,0,1,0,0,3,0,0,1,0,0,0,0,0},
														{0,0,0,0,1,2,1,0,1,0,1,2,1,0,0,0,0},
														{0,0,0,0,1,2,1,0,1,0,1,2,1,0,0,0,0},
														{0,0,0,0,0,1,1,1,1,1,1,1,0,0,0,0,0},
														{0,0,0,0,0,1,1,2,1,2,1,1,0,0,0,0,0},
														{0,0,0,3,0,0,1,1,2,1,1,0,0,3,0,0,0},
														{0,0,0,3,0,0,0,1,2,1,0,0,0,3,0,0,0},
														{0,0,1,1,1,0,0,1,2,1,0,0,1,1,1,0,0},
														{0,1,1,1,1,1,0,1,2,1,0,1,1,1,1,1,0},
														{0,1,1,1,1,1,0,1,2,1,0,1,1,1,1,1,0},
														{0,1,1,1,1,1,1,1,2,1,1,1,1,1,1,1,0},
														{0,1,1,2,1,1,1,1,1,1,1,1,1,2,1,1,0},
														{0,1,1,2,1,1,1,0,0,0,1,1,1,2,1,1,0},
														{0,0,1,2,1,0,0,0,0,0,0,0,1,2,1,0,0},
														{0,0,0,1,0,0,0,0,0,0,0,0,0,1,0,0,0} };
														
																											
static int8_t bullet_shape[4][3] = { 	{0,1,0},
										{1,2,1},
										{1,2,1},
										{0,1,0} };
										
static int8_t bomb_shape[8][5] = { 		{1,0,0,0,1},
										{0,1,1,1,0},
										{0,0,1,0,0},
										{0,1,1,1,0},
										{1,1,1,1,1},
										{1,1,1,1,1},
										{0,1,1,1,0},
										{0,0,1,0,0} };
typedef struct{
	int x, y;
} position;
													
static int8_t invader_matrix[LINE][COL];
static int8_t invader_direction = 1;
static int8_t boss_direction = 1;
static int8_t invader_mov_start = START_INV_Y;
static int8_t ship_direction;
static int8_t ship_y;
static int8_t ship_speed = 10;
static position bullets[15];
static int8_t bullet_num;
static int8_t bullet_speed = 5;
static int8_t invaders_num;
static position bombs[15];
static int8_t bomb_num;
static int8_t bomb_speed = 4;
static int8_t bomb_rate;
static int8_t health;
static int8_t boss_health;
static int8_t boss_y;
static int8_t shield = 0;
static int8_t bomb_drop, bomb_min_range, bomb_max_range;
static int8_t bullet_height;

void init_invaders(){
	int8_t i, j;
	for(i = 0; i < INVADER_LN ; i++)
		for(j = 0; j < INVADER_COL; j++)
			invader_matrix[i][j] = 1;
}

void buttons_init(){
	/* RIGHT */
	DDRC &= ~(1 << PC0);
	PORTC |= (1 << PC0);
	
	/* DOWN */
	DDRC &= ~(1 << PC1);
	PORTC |= (1 << PC1);
	
	/* LEFT */
	DDRC &= ~(1 << PC2);
	PORTC |= (1 << PC2);
	
	/* UP */
	DDRC &= ~(1 << PC3);
	PORTC |= (1 << PC3);
	
	DDRD |= (1 << PD4);
	PORTD &= ~(1 << PD4);
}

#define BUZZER_PORT PORTD
#define BUZZER_DDR  DDRD
#define BUZZER_PIN  4

void BUZZ(float duration, float period)
{
    int i,cycles;
    float half_period;  // Initialize variables
 
    cycles=duration/period; // Compute the number of cycles to loop toggling the pin
    half_period = period/2; // Compute a half cycle period
 
    BUZZER_DDR = (1 << BUZZER_PIN) | BUZZER_DDR;  // Set the port for the buzzer output
 
    for (i=0;i<cycles;i++)   // Toggle the speaker the appropriate number of cycles
    {
        _delay_ms(half_period);                         // Wait a half cycle to toggle port pin
        BUZZER_PORT = (1 << BUZZER_PIN) | BUZZER_PORT;    // Set the port pin
        _delay_ms(half_period);                         // Wait a half cycle to clear the port pin
        BUZZER_PORT = ~(1 << BUZZER_PIN) & BUZZER_PORT;   // Clear the port pin
    }
 
    return;     // Return to the main program
}
void draw_invader(int8_t core_x, int8_t core_y){
	int8_t x, y;
	for(x = -INVADER_SIZE/2 ;x <= INVADER_SIZE/2 -2; x++)
		for(y = -INVADER_SIZE/2 ;y <= INVADER_SIZE/2; y++)
			if(invader_shape[(x+INVADER_SIZE/2) * INVADER_SIZE
				+ (y+INVADER_SIZE/2)] > 0)
				LCD_draw_pixel(core_y + y, core_x + x, ST7735_GREEN);
			else
				LCD_draw_pixel(core_y + y, core_x + x, ST7735_BLACK);

}

void draw_ship(int8_t core_x, int8_t core_y){
	int8_t x, y, pix;
	for(x = -SHIP_HEIGHT/2 ;x <= SHIP_HEIGHT/2 ; x++)
		for(y = -SHIP_WIDTH/2 ;y <= SHIP_WIDTH/2; y++){
			pix = ship_shape[x + SHIP_HEIGHT/2][y + SHIP_WIDTH/2];
			if( pix == 0)
				LCD_draw_pixel(core_y + y, core_x + x, ST7735_BLACK);
			else
				if(pix == 1)
					LCD_draw_pixel(core_y + y, core_x + x, ST7735_BLUE);
				else
				if(pix == 2)
					LCD_draw_pixel(core_y + y, core_x + x, ST7735_CYAN);
				else
					LCD_draw_pixel(core_y + y, core_x + x, ST7735_RED);
		}
}

void draw_invaders(int8_t x, int8_t y){
	int8_t i,j, start;
	start = y;
	for(i = 0 ;i< INVADER_LN; i++){
		for(j = 0; j<INVADER_COL; j++){
			if(invader_matrix[i][j] > 0)
				draw_invader(x, y);
			y += INVADER_SIZE + 1;
		}
		y = start;
		x += INVADER_SIZE;
	}
}

void draw_bullet(int bullet){
	int16_t i,j;
	int x, y;
	x = bullets[bullet].x;
	y = bullets[bullet].y;
	for(i = 0 ; i<4; i++)
		for(j = 0; j<3; j++)
			if(bullet_shape[i][j] == 0)
				LCD_draw_pixel(y + j - 1, x - i, ST7735_BLACK);
			else
				if(bullet_shape[i][j] == 1)
					LCD_draw_pixel(y + j - 1, x - i, ST7735_YELLOW);
				else
					LCD_draw_pixel(y + j - 1, x - i, ST7735_RED);

}
void draw_bomb(int bomb){
	int8_t i,j;
	int8_t x, y;
	x = bombs[bomb].x;
	y = bombs[bomb].y;
	for(i = 0 ; i<8; i++)
		for(j = 0; j<5; j++)
			if(bomb_shape[i][j] == 0)
				LCD_draw_pixel(y + j - 2, x + i, ST7735_BLACK);
			else
				LCD_draw_pixel(y + j - 2, x + i, 0x001F);
}

void draw_health(){
	int8_t i, j;
	for( i =0 ; i<4; i++)
		for(j = 0; j< 4 * START_HEALTH; j++)
			if(j < health * 4)
				LCD_draw_pixel(WIDTH/2 + j - START_HEALTH*2, 5 + i, ST7735_RED);
			else
				LCD_draw_pixel(WIDTH/2 + j - START_HEALTH*2, 5 + i, ST7735_BLACK);
}
void draw_boss_health(){
	int8_t i, j;
	for( i =0 ; i<4; i++)
		for(j = 0; j< 100; j++)
			if(j < boss_health * 2)
				LCD_draw_pixel(WIDTH/2 + j - 50, 12 + i, ST7735_BLUE);
			else
				LCD_draw_pixel(WIDTH/2 + j - 50, 12 + i, ST7735_BLACK);
}

void move_invaders(){
	if(invader_mov_start + (INVADER_COL-1)*(INVADER_SIZE + 1) == WIDTH - INVADER_SIZE/2)
		invader_direction = -1;
	if(invader_mov_start - INVADER_SIZE/2 == WIDTH_OFF)
		invader_direction = 1;
	draw_invaders(START_INV_X, invader_mov_start += invader_direction);
}

void clean_up_boss(){
	int i,j;
	if(boss_direction > 0)
	for( i = 0; i<BOSS_HEIGHT; i++)
		for(j = 0; j<=BOSS_SPEED;j++)
			LCD_draw_pixel(boss_y - j +1, i + 22, ST7735_BLACK);
	else
	for( i = 0; i<=BOSS_HEIGHT; i++)
		for(j = 0; j<=BOSS_SPEED;j++)
			LCD_draw_pixel(boss_y + BOSS_WIDTH + j-1, i + 22, ST7735_BLACK);
}			


void move_boss(){
	if(boss_y < WIDTH_OFF + BOSS_SPEED)
		boss_direction = 1;
	if(boss_y + BOSS_WIDTH > WIDTH - BOSS_SPEED)
		boss_direction = -1;
	boss_y += BOSS_SPEED * boss_direction;
	draw_boss(boss_y);
	clean_up_boss();
}

void clean_up_trail(){
	int8_t x, y;
	if(ship_direction > 0)
	for( x = -SHIP_HEIGHT/2 ;x <= SHIP_HEIGHT/2 ; x++)
		for(y = 0; y <= ship_speed; y++)
			LCD_draw_pixel(ship_y - SHIP_WIDTH/2 - y, SHIP_START_X + x, ST7735_BLACK);
	else
		for( x = -SHIP_HEIGHT/2 ;x <= SHIP_HEIGHT/2 ; x++)
			for(y = 0; y <= ship_speed; y++)
				LCD_draw_pixel(ship_y + SHIP_WIDTH/2 + y, SHIP_START_X + x, ST7735_BLACK);
}
void move_ship(){
	if(ship_direction != 0){
		int8_t new_pos;
		new_pos = ship_y + ship_speed*ship_direction;
		if((new_pos >= WIDTH_OFF + SHIP_WIDTH/2) && (new_pos <= WIDTH - SHIP_WIDTH/2+1)){
			ship_y = new_pos;
			draw_ship(SHIP_START_X, ship_y);
			clean_up_trail();
		}
	}
}

void shoot(){
	bullets[bullet_num].x = SHIP_START_X - SHIP_HEIGHT/2 - 2;
	bullets[bullet_num].y = ship_y;
	draw_bullet(bullet_num);
	bullet_num ++;
}
void invader_bomb(){
	bombs[bomb_num].x = bomb_drop;
	bombs[bomb_num].y = bomb_min_range + rand() %(bomb_max_range - bomb_min_range);
	draw_bomb(bomb_num);
	bomb_num ++;
}

void clean_up_bullet(int8_t bullet){
	int8_t i, j;
	for(i = 0 ; i<bullet_speed; i++)
		for(j = 0; j<3; j++)
				LCD_draw_pixel(bullets[bullet].y + j - 1, bullets[bullet].x + i + 1, ST7735_BLACK);
}

void destroy_bullet(int8_t bullet){
	int8_t i,j;
	for(i = 0 ; i<4; i++)
		for(j = 0; j<3; j++)
				LCD_draw_pixel(bullets[bullet].y + j - 1, bullets[bullet].x - i, ST7735_BLACK);
	bullets[bullet] = bullets[bullet_num-1];
	bullet_num--;
}

void destroy_bomb(int8_t bomb){
	int8_t i,j;
	for(i = 0 ; i<8; i++)
		for(j = 0; j<5; j++)
				LCD_draw_pixel(bombs[bomb].y + j - 2, bombs[bomb].x + i, ST7735_BLACK);
	bombs[bomb] = bombs[bomb_num-1];
	bomb_num--;
}

void move_bullets(){
	int i;
	for(i = 0; i < bullet_num; i++){
		bullets[i].x -= bullet_speed;
		if(bullets[i].x < bullet_height){
			clean_up_bullet(i);
			destroy_bullet(i);
			i--;
		}
		else{
			draw_bullet(i);
			clean_up_bullet(i);
		}
	}		
}

void destroy_invader(int8_t x, int8_t y){
	int8_t i, j;
	for(i= 0; i< INVADER_SIZE -2; i++)
		for(j = 0 ; j< INVADER_SIZE; j++)
			LCD_draw_pixel(y+j,x+i , ST7735_BLACK);
}
void clean_up_bomb(int8_t bomb){
	int8_t i, j;
	for(i = 0 ; i<bomb_speed; i++)
		for(j = 0; j<5; j++)
				LCD_draw_pixel(bombs[bomb].y + j -2, bombs[bomb].x - i - 1, ST7735_BLACK);
}

void move_bombs(){
	int i;
	for(i = 0; i < bomb_num; i++){
		bombs[i].x += bomb_speed;
		if(bombs[i].x > HEIGHT){
			clean_up_bomb(i);
			destroy_bomb(i);
			i--;
		}
		else{
			draw_bomb(i);
			clean_up_bomb(i);
		}
	}
}
void collision_detection(){
	int8_t i, j, k;
	int8_t invaders_x, invaders_y;
	invaders_x = START_INV_X;
	invaders_y = invader_mov_start;
	for(i = 0 ;i< INVADER_LN; i++){
		for(j = 0; j<INVADER_COL; j++){
			for( k = 0; k<bullet_num; k++){
				if(bullets[k].x < HEIGHT/2){
					if((bullets[k].x - 4 < invaders_x + INVADER_SIZE/2 - 1) &&
						(bullets[k].y > invaders_y - INVADER_SIZE/2 +1) &&
						(bullets[k].y < invaders_y + INVADER_SIZE/2 -1) &&
						invader_matrix[i][j] >0){
						destroy_bullet(k);
						destroy_invader(invaders_x - 5, invaders_y - 5);
						invaders_num--;
						invader_matrix[i][j] = 0;
						}
				}
			}
			invaders_y += INVADER_SIZE + 1;
		}
		invaders_y = invader_mov_start;
		invaders_x += INVADER_SIZE;
	}
	for(i = 0 ;i< bomb_num; i++){
		if((bombs[i].x + 8 > SHIP_START_X - SHIP_HEIGHT/2) &&
			(bombs[i].x < SHIP_START_X + SHIP_HEIGHT/2) &&
			(bombs[i].y > ship_y - SHIP_WIDTH/2) &&
			(bombs[i].y < ship_y + SHIP_WIDTH/2)){
				
			destroy_bomb(i);
			draw_ship(SHIP_START_X, ship_y);
			if(shield > 0)
				health -= 1;
			else	
				health -= 2;
			BUZZ(600,1);
			BUZZ(400,1);
			draw_health();
		}
	}
	draw_health();
}

void boss_collision_detection(){
	int8_t i, k;
	int8_t invaders_x, invaders_y;
	invaders_x = START_INV_X;
	invaders_y = invader_mov_start;
	for( k = 0; k<bullet_num; k++){
		if(bullets[k].x < HEIGHT/2){
			if((bullets[k].x - 4 < 53) &&
				(bullets[k].y > boss_y) &&
				(bullets[k].y < boss_y + 51) &&
				(bullets[k].x - 4 > 30)){
					
				destroy_bullet(k);
				boss_health -= 1;
				draw_boss_health();
			}
		}
	}
	for(i = 0 ;i< bomb_num; i++){
		if((bombs[i].x + 8 > SHIP_START_X - SHIP_HEIGHT/2) &&
			(bombs[i].x < SHIP_START_X + SHIP_HEIGHT/2) &&
			(bombs[i].y > ship_y - SHIP_WIDTH/2) &&
			(bombs[i].y < ship_y + SHIP_WIDTH/2)){
				
			destroy_bomb(i);
			draw_ship(SHIP_START_X, ship_y);
			if(shield > 0)
				health -= 1;
			else	
				health -= 2;
			BUZZ(600,1);
			BUZZ(400,1);
			draw_health();
		}
	}
}

void play_game(){
	LCD_fill_screen(ST7735_BLACK);
	
	GFX_set_text_color(ST7735_RED);
	GFX_set_cursor(25,30);
	GFX_print_str("SPACE INVADERS");
	_delay_ms(75000);
	while((PINC & (1 << PC2)) != 0){}
	
	LCD_fill_screen(ST7735_BLACK);
	
	INVADER_LN = LINE;
	INVADER_COL = COL;
	invaders_num = INVADER_LN * INVADER_COL;
	init_invaders();
	
	ship_y = SHIP_START_Y;
	draw_ship(SHIP_START_X, SHIP_START_Y);
	
	int8_t shot = 0, bombed = 0,i,j, level = 1, shield;
	bullet_num = 0;
	bomb_num = 0;
	health = START_HEALTH;
	boss_health = 50;
	draw_health();
	bomb_drop = 20;
	bomb_rate = 20;
	bomb_max_range = 125;
	bomb_min_range = 5;
	bullet_height = 20;
	while(1){
		shield = 0;
		shot ++;
		bombed ++;
		move_invaders();
		ship_direction = 0;
		if ((PINC & (1 << PC0)) == 0){
			ship_direction = -1;
		}
		if ((PINC & (1 << PC1)) == 0){
			ship_direction = 1;
		}
		move_ship();
		if ((PINC & (1 << PC2)) == 0 && bullet_num < 10 && shot >= 5){
			shoot();
			BUZZ(100,1);
			shot = 0;
		}
		if ((PINC & (1 << PC3)) == 0){
			shield = 1;
		}
		move_bullets();
		move_bombs();
		collision_detection();
		if(bombed >= bomb_rate){
			bombed = 0;
			invader_bomb();
		}
		if(invaders_num == 0){
			if(level == 1){
				level = 2;
				INVADER_LN = 2;
				init_invaders();
				invaders_num = INVADER_COL * INVADER_LN;
				bomb_rate = 10;
				LCD_fill_screen(ST7735_BLACK);
				GFX_set_cursor(30,55);
				GFX_print_str("LEVEL 2");
				GFX_set_cursor(25,65);
				GFX_print_str("MORE SPEED,");
				GFX_set_cursor(25,75);
				GFX_print_str("MORE BOMBS");
				for(j = 0; j<15; j++)
					for(i = 0; i<10 ; i++)
						BUZZ(j * 25,20);
				LCD_fill_screen(ST7735_BLACK);
				bullet_num = 0;
				bomb_num = 0;
			}
			else 
				break;
		}
		if(health <= 0){ 
			lost_game();
			break;
		}
		if(level == 1)
			_delay_ms((INVADER_COL*INVADER_LN - invaders_num)*35);
	}
	
	GFX_set_cursor(15,63);
	GFX_set_text_size(1);
	LCD_fill_screen(ST7735_BLACK);
	GFX_print_str("PREPARE YOURSELF");
	for(j = 0; j<25; j++)
	for(i = 0; i<10 ; i++)
		BUZZ(i * 50,20);
	_delay_ms(750000);
	_delay_ms(750000);
	_delay_ms(750000);
	LCD_fill_screen(ST7735_BLACK);
	ship_y = SHIP_START_Y;
	draw_ship(SHIP_START_X, ship_y);
	boss_y = WIDTH/2 - BOSS_WIDTH/2;
	bullet_num = 0;
	bomb_num = 0;
	draw_boss(boss_y);
	draw_boss_health();
	bomb_drop = 50;
	bullet_height = 40;
	draw_health();
	while(1){
		bomb_max_range = boss_y + BOSS_WIDTH;
		bomb_min_range = boss_y;
		move_boss();
		shot ++;
		bombed ++;
		move_invaders();
		ship_direction = 0;
		if ((PINC & (1 << PC0)) == 0){
			ship_direction = -1;
		}
		if ((PINC & (1 << PC1)) == 0){
			ship_direction = 1;
		}
		move_ship();
		if ((PINC & (1 << PC2)) == 0 && bullet_num < 10 && shot >= 5){
			shoot();
			BUZZ(100,1);
			shot = 0;
		}
		move_bullets();
		move_bombs();
		boss_collision_detection();
		if(bombed >= 10){
			bombed = 0;
			invader_bomb();
		}
		if(boss_health <= 0)
			won_game();
		if(health <= 0)
			lost_game();	
	}
}

void won_game(){
	BUZZ(600,1);
	BUZZ(700,1);	
	LCD_fill_screen(ST7735_CYAN);
	GFX_set_text_color(ST7735_BLUE);
	GFX_set_cursor(15,30);
	GFX_print_str("CONGRATULATIONS!");
	GFX_set_cursor(15,40);
	GFX_print_str("YOU HAVE DEFEATED");
	GFX_set_cursor(30,50);
	GFX_print_str("THE INVASION");
	GFX_set_cursor(15,80);
	GFX_print_str("PRESS 'SHOOT' TO");
	GFX_set_cursor(35,90);
	GFX_print_str("PLAY AGAIN");	
	_delay_ms(750000);
	_delay_ms(750000);
	_delay_ms(750000);	
	while (1) {		
		if ((PINC & (1 << PC2)) == 0)
			play_game();
	} 
}
void lost_game(){
	LCD_fill_screen(ST7735_RED);
	BUZZ(600,1);
	BUZZ(500,1);
	GFX_set_text_color(ST7735_BLUE);
	GFX_set_cursor(45,30);
	GFX_print_str("DEFEAT!");
	GFX_set_cursor(6,40);
	GFX_print_str("YOUR PLANET HAS BEEN");
	GFX_set_cursor(43,50);
	GFX_print_str("INVADED!");
	GFX_set_cursor(15,80);
	GFX_print_str("PRESS 'SHOOT' TO");
	GFX_set_cursor(35,90);
	GFX_print_str("PLAY AGAIN");
	_delay_ms(750000);
	_delay_ms(750000);
	_delay_ms(750000);
	while (1) {
		if ((PINC & (1 << PC2)) == 0)
			play_game();
	} 
}

int main(void){
	buttons_init();
	LCD_init();
	GFX_init(130,131);
	play_game();
    return 0;
}




